{% load crispy_forms_field %}

{% if field.is_hidden %}
	{{ field }}
{% else %}
    {% if field|is_checkbox %}
        <div class="form-group">
    {% endif %}
	<{% if tag %}{{ tag }}{% else %}div{% endif %} id="div_{{ field.auto_id }}" class="form-group{% if wrapper_class %} {{ wrapper_class }}{% endif %}{% if form_show_errors%}{% if field.errors %} has-error{% endif %}{% endif %}{% if field.css_classes %} {{ field.css_classes }}{% endif %}">
		{% if field.label and not field|is_checkbox and form_show_labels %}
			<label for="{{ field.id_for_label }}" class="control-label {{ label_class }}{% if field.field.required %} requiredField{% endif %}">
				{{ field.label|safe }}{% if field.field.required %}<span class="asteriskField">*</span>{% endif %}
			</label>
		{% endif %}

        {% if field|is_checkboxselectmultiple %}
            {% include 'bootstrap3/layout/checkboxselectmultiple.html' %}
        {% endif %}

        {% if field|is_radioselect %}
            {% include 'bootstrap3/layout/radioselect.html' %}
        {% endif %}

        {% if not field|is_checkboxselectmultiple and not field|is_radioselect %}
            {% if field|is_checkbox and form_show_labels %}
                {% if label_class %}
                    <div class="controls col-lg-offset-{{ label_size }} {{ field_class }}">
                {% endif %}
                <label for="{{ field.id_for_label }}" class="{% if field.field.required %} requiredField{% endif %}">
                    {% crispy_field field 'class' 'checkbox' %}
                    {{ field.label|safe }}
                    {% include 'bootstrap3/layout/help_text_and_errors.html' %}
                </label>
                {% if label_class %}
                    </div>
                {% endif %}
            {% else %}
                <div class="controls {{ field_class }}">
                    {% crispy_field field %}
                    {% include 'bootstrap3/layout/help_text_and_errors.html' %}
                </div>
            {% endif %}
        {% endif %}
	</{% if tag %}{{ tag }}{% else %}div{% endif %}>
    {% if field|is_checkbox %}
        </div>
    {% endif %}
{% endif %}
